import os
import abc
import json
import time
from datetime import datetime, timezone

import appdirs  # type: ignore

from biolib.biolib_errors import BioLibError
from biolib.biolib_logging import logger_no_user_data
from biolib.typing_utils import Optional, Generic, TypeVar

StateType = TypeVar('StateType')  # pylint: disable=invalid-name


class CacheStateError(BioLibError):
    pass


class CacheState(abc.ABC, Generic[StateType]):
    @property
    @abc.abstractmethod
    def _state_path(self) -> str:
        raise NotImplementedError

    @abc.abstractmethod
    def _get_default_state(self) -> StateType:
        raise NotImplementedError

    @property
    def _user_cache_dir(self) -> str:
        user_cache_dir: str = appdirs.user_cache_dir(appname='pybiolib', appauthor='biolib')
        os.makedirs(user_cache_dir, exist_ok=True)
        return user_cache_dir

    @property
    def _state_lock_path(self) -> str:
        return f'{self._state_path}.lock'

    def __init__(self) -> None:
        self._state: Optional[StateType] = None

    def __enter__(self) -> StateType:
        logger_no_user_data.debug(f'CacheState: Entering state path: {self._state_path}...')
        try:
            self._acquire_state_lock()
            if os.path.exists(self._state_path):
                with open(self._state_path, mode='r') as file:
                    self._state = json.loads(file.read())
            else:
                self._state = self._get_default_state()
                with open(self._state_path, mode='w') as file:
                    file.write(json.dumps(self._state))

            # Check for type checking
            if self._state is None:
                raise CacheStateError('Internal state is not defined')
        except BaseException as error:  # pylint: disable=broad-except
            logger_no_user_data.debug(f'Could not get LFS lock, got error: {error}...')
            raise error
        return self._state

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        with open(self._state_path, mode='w') as file:
            file.write(json.dumps(self._state))

        self._release_state_lock()
        logger_no_user_data.debug(f'CacheState: Exited state path: {self._state_path}')

    def _acquire_state_lock(self) -> None:
        for _ in range(10):
            try:
                lock_file = open(self._state_lock_path, mode='x')
                lock_file.close()
                return
            except BaseException as error:  # pylint: disable=broad-except
                logger_no_user_data.debug(f'Failed to acquire lock file "{self._state_lock_path}". Got error: {error}')

            time.sleep(0.5)

        raise CacheStateError(f'Cache state timed out waiting to acquire lock file "{self._state_lock_path}"')

    def _release_state_lock(self) -> None:
        if os.path.exists(self._state_lock_path):
            os.remove(self._state_lock_path)
        else:
            raise CacheStateError('Cache state was not locked.')

    @staticmethod
    def get_timestamp_now() -> str:
        return datetime.now(timezone.utc).isoformat()



import time
from pathlib import Path

from biolib._internal.utils import PathFilter, filter_lazy_loaded_files
from biolib.biolib_binary_format import ModuleOutputV2
from biolib.biolib_binary_format.remote_endpoints import RemoteJobStorageEndpoint
from biolib.biolib_binary_format.remote_stream_seeker import StreamSeeker
from biolib.biolib_binary_format.utils import LazyLoadedFile, RemoteIndexableBuffer
from biolib.biolib_errors import BioLibError
from biolib.biolib_logging import logger
from biolib.typing_utils import Dict, List, Optional


class JobResult:
    def __init__(
        self,
        job_uuid: str,
        job_auth_token: str,
        module_output: Optional[ModuleOutputV2] = None,
    ):
        self._job_uuid: str = job_uuid
        self._job_auth_token: str = job_auth_token

        self._module_output: Optional[ModuleOutputV2] = module_output

    def get_stdout(self) -> bytes:
        return self._get_module_output().get_stdout()

    def get_stderr(self) -> bytes:
        return self._get_module_output().get_stderr()

    def get_exit_code(self) -> int:
        return self._get_module_output().get_exit_code()

    def save_files(
        self,
        output_dir: str,
        path_filter: Optional[PathFilter] = None,
        skip_file_if_exists: bool = False,
        overwrite: bool = False,
        flat: bool = False,
    ) -> None:
        module_output = self._get_module_output()
        output_files = module_output.get_files()
        filtered_output_files = filter_lazy_loaded_files(output_files, path_filter) if path_filter else output_files

        if len(filtered_output_files) == 0:
            logger.debug('No output files to save')
            return

        if flat:
            basename_to_paths: Dict[str, List[str]] = {}
            for file in filtered_output_files:
                basename = Path(file.path).name
                if basename not in basename_to_paths:
                    basename_to_paths[basename] = []
                basename_to_paths[basename].append(file.path)

            duplicates = {basename: paths for basename, paths in basename_to_paths.items() if len(paths) > 1}

            if duplicates:
                max_shown = 3
                error_parts = []
                sorted_basenames = sorted(duplicates.keys())

                for basename in sorted_basenames[:max_shown]:
                    paths = duplicates[basename]
                    error_parts.append(f'  {basename}: ({", ".join(paths)})')

                error_message = 'Cannot save files in flat mode: duplicate filenames detected:\n' + '\n'.join(
                    error_parts
                )

                if len(duplicates) > max_shown:
                    remaining = len(duplicates) - max_shown
                    error_message += f'\n  (and {remaining} more)'

                raise BioLibError(error_message)

        major_gap_threshold = 50_000
        n = len(filtered_output_files)

        next_break_end = [0] * n
        if n > 0:
            next_break_end[n - 1] = filtered_output_files[n - 1].start + filtered_output_files[n - 1].length
            for i in range(n - 2, -1, -1):
                end_i = filtered_output_files[i].start + filtered_output_files[i].length
                gap = filtered_output_files[i + 1].start - end_i
                if gap >= major_gap_threshold:
                    next_break_end[i] = end_i
                else:
                    next_break_end[i] = next_break_end[i + 1]

        total_files_data_to_download_in_bytes = 0
        file_read_ahead_map = {}
        for i, file in enumerate(filtered_output_files):
            total_files_data_to_download_in_bytes += file.length
            end_i = file.start + file.length
            read_ahead_bytes = max(0, next_break_end[i] - end_i)
            file_read_ahead_map[i] = read_ahead_bytes

        # Assume files are in order
        first_file = filtered_output_files[0]
        last_file = filtered_output_files[len(filtered_output_files) - 1]
        stream_seeker = StreamSeeker(
            files_data_start=first_file.start,
            files_data_end=last_file.start + last_file.length,
            max_chunk_size=min(total_files_data_to_download_in_bytes, 10_000_000),
            upstream_buffer=module_output.buffer,
        )

        logger.info(f'Saving {len(filtered_output_files)} files to {output_dir}...')
        for file_index, file in enumerate(filtered_output_files):
            if flat:
                destination_file_path = Path(output_dir) / Path(file.path).name
            else:
                # Remove leading slash of file_path
                destination_file_path = Path(output_dir) / Path(file.path.lstrip('/'))

            if destination_file_path.exists():
                if skip_file_if_exists:
                    print(f'Skipping {destination_file_path} as a file with that name already exists locally.')
                    continue
                elif not overwrite:
                    raise BioLibError(f'File {destination_file_path} already exists. Set overwrite=True to overwrite.')
                else:
                    destination_file_path.rename(
                        f'{destination_file_path}.biolib-renamed.{time.strftime("%Y%m%d%H%M%S")}'
                    )

            dir_path = destination_file_path.parent
            if dir_path:
                dir_path.mkdir(parents=True, exist_ok=True)

            # write content to temporary (partial) file
            partial_path = destination_file_path.with_suffix(
                destination_file_path.suffix + f'.{self._job_uuid}.partial_biolib_download'
            )
            file_start = file.start
            data_to_download = file.length
            if partial_path.exists():
                data_already_downloaded = partial_path.stat().st_size
                file_start += data_already_downloaded
                data_to_download -= data_already_downloaded

            read_ahead_bytes = file_read_ahead_map[file_index]

            with open(partial_path, mode='ab') as partial_file:
                for chunk in stream_seeker.seek_and_read(
                    file_start=file_start, file_length=data_to_download, read_ahead_bytes=read_ahead_bytes
                ):
                    partial_file.write(chunk)

            # rename partial file to actual file name
            partial_path.rename(destination_file_path)

    def get_output_file(self, filename) -> LazyLoadedFile:
        files = self._get_module_output().get_files()
        filtered_files = filter_lazy_loaded_files(files, path_filter=filename)
        if not filtered_files:
            raise BioLibError(f'File {filename} not found in results.')

        if len(filtered_files) != 1:
            raise BioLibError(f'Found multiple results for filename {filename}.')

        return filtered_files[0]

    def list_output_files(self, path_filter: Optional[PathFilter] = None) -> List[LazyLoadedFile]:
        files = self._get_module_output().get_files()
        if not path_filter:
            return files

        return filter_lazy_loaded_files(files, path_filter)

    def _get_module_output(self) -> ModuleOutputV2:
        if self._module_output is None:
            remote_job_storage_endpoint = RemoteJobStorageEndpoint(
                job_auth_token=self._job_auth_token,
                job_uuid=self._job_uuid,
                storage_type='output',
            )
            buffer = RemoteIndexableBuffer(endpoint=remote_job_storage_endpoint)
            self._module_output = ModuleOutputV2(buffer)

        return self._module_output



from __future__ import annotations

import os
from typing import Callable, Iterator

import biolib.api as api
from biolib._internal.file_utils import get_files_and_size_of_directory, get_iterable_zip_stream
from biolib.biolib_errors import BioLibError
from biolib.biolib_logging import logger
from biolib.typing_utils import List, Optional, Tuple
from biolib.utils import MultiPartUploader


def _upload_from_iterator(
    payload_iterator: Iterator[bytes],
    payload_size_in_bytes: int,
    resource_uuid: Optional[str] = None,
    resource_version_uuid: Optional[str] = None,
    use_process_pool: bool = False,
    publish: bool = False,
    on_progress: Optional[Callable[[int, int], None]] = None,
) -> str:
    if (resource_uuid is None) == (resource_version_uuid is None):
        raise ValueError('Must provide exactly one of resource_uuid or resource_version_uuid')

    if resource_version_uuid is None:
        response = api.client.post(
            path='/lfs/versions/',
            data={'resource_uuid': resource_uuid},
        )
        resource_version_uuid = response.json()['uuid']

    multipart_uploader = MultiPartUploader(
        use_process_pool=use_process_pool,
        get_presigned_upload_url_request={
            'headers': None,
            'requires_biolib_auth': True,
            'path': f'/lfs/versions/{resource_version_uuid}/presigned_upload_url/',
        },
        complete_upload_request={
            'headers': None,
            'requires_biolib_auth': True,
            'path': f'/lfs/versions/{resource_version_uuid}/complete_upload/',
        },
        on_progress=on_progress,
    )
    multipart_uploader.upload(payload_iterator=payload_iterator, payload_size_in_bytes=payload_size_in_bytes)

    if publish:
        api.client.patch(
            path=f'/resources/versions/{resource_version_uuid}/',
            data={'state': 'published', 'set_as_active': True},
        )

    return resource_version_uuid


def validate_data_path_and_get_files_and_size_of_directory(data_path: str) -> Tuple[List[str], int]:
    assert os.path.isdir(data_path), f'The path "{data_path}" is not a directory.'

    if os.path.realpath(data_path) == '/':
        raise BioLibError('Pushing your root directory is not possible')

    original_working_dir = os.getcwd()
    os.chdir(data_path)
    files_to_zip, data_size_in_bytes = get_files_and_size_of_directory(directory=os.getcwd())
    os.chdir(original_working_dir)

    if data_size_in_bytes > 4_500_000_000_000:
        raise BioLibError('Attempted to push directory with a size larger than the limit of 4.5 TB')

    return files_to_zip, data_size_in_bytes


def push_data_path(
    data_path: str,
    data_size_in_bytes: int,
    files_to_zip: List[str],
    resource_uuid: Optional[str] = None,
    resource_version_uuid: Optional[str] = None,
    chunk_size_in_mb: Optional[int] = None,
    publish: bool = False,
) -> str:
    if (resource_uuid is None) == (resource_version_uuid is None):
        raise ValueError('Must provide exactly one of resource_uuid or resource_version_uuid')

    original_working_dir = os.getcwd()
    os.chdir(data_path)

    min_chunk_size_bytes = 10_000_000
    chunk_size_in_bytes: int
    if chunk_size_in_mb:
        chunk_size_in_bytes = chunk_size_in_mb * 1_000_000  # Convert megabytes to bytes
        if chunk_size_in_bytes < min_chunk_size_bytes:
            logger.warning('Specified chunk size is too small, using minimum of 10 MB instead.')
            chunk_size_in_bytes = min_chunk_size_bytes
    else:
        # Calculate chunk size based on max chunk count of 10_000, using 9_000 to be on the safe side
        chunk_size_in_bytes = max(min_chunk_size_bytes, int(data_size_in_bytes / 9_000))

    data_size_in_mb = round(data_size_in_bytes / 10**6)
    logger.info(f'Zipping {len(files_to_zip)} files, in total ~{data_size_in_mb}mb of data')

    iterable_zip_stream = get_iterable_zip_stream(files=files_to_zip, chunk_size=chunk_size_in_bytes)

    new_resource_version_uuid = _upload_from_iterator(
        payload_iterator=iterable_zip_stream,
        payload_size_in_bytes=data_size_in_bytes,
        resource_uuid=resource_uuid,
        resource_version_uuid=resource_version_uuid,
        use_process_pool=True,
        publish=publish,
    )

    os.chdir(original_working_dir)
    return new_resource_version_uuid



import re


def normalize_for_docker_tag(name: str) -> str:
    if not name:
        return ''

    normalized = re.sub(r'[^a-z0-9-]', '-', name.lower())

    normalized = re.sub(r'-+', '-', normalized)
    normalized = normalized.strip('-')

    return normalized



from biolib.biolib_binary_format.utils import IndexableBuffer
from biolib.biolib_logging import logger
from biolib.typing_utils import Iterable


class StreamSeeker:
    def __init__(
        self,
        upstream_buffer: IndexableBuffer,
        files_data_start: int,
        files_data_end: int,
        max_chunk_size: int,
    ):
        self._upstream_buffer = upstream_buffer
        self._files_data_end = files_data_end
        self._max_chunk_size = max_chunk_size

        self._buffer_start = files_data_start
        self._buffer = bytearray()

    def seek_and_read(self, file_start: int, file_length: int, read_ahead_bytes: int = 0) -> Iterable[bytes]:
        assert file_start >= self._buffer_start
        self._buffer = self._buffer[file_start - self._buffer_start :]
        self._buffer_start = file_start

        while True:
            file_byte_count_remaining = file_length - (self._buffer_start - file_start)
            if file_byte_count_remaining <= 0:
                return

            if len(self._buffer) > 0:
                take = min(file_byte_count_remaining, len(self._buffer))
                chunk = self._buffer[:take]
                if chunk:
                    yield chunk
                self._buffer = self._buffer[take:]
                self._buffer_start += take
            else:
                start_of_fetch = self._buffer_start + len(self._buffer)
                bytes_left_in_stream = self._files_data_end - start_of_fetch
                if bytes_left_in_stream <= 0:
                    logger.error(
                        'StreamSeeker: no bytes left upstream (start_of_fetch=%d, files_data_end=%d)',
                        start_of_fetch,
                        self._files_data_end,
                    )
                    return

                fetch_size = min(self._max_chunk_size, file_byte_count_remaining + read_ahead_bytes)
                if fetch_size > bytes_left_in_stream:
                    logger.error(
                        'StreamSeeker: fetch_size (%d) > bytes_left_in_stream (%d); clamping',
                        fetch_size,
                        bytes_left_in_stream,
                    )
                    fetch_size = bytes_left_in_stream

                fetched_data = self._upstream_buffer.get_data(start=start_of_fetch, length=fetch_size)
                self._buffer.extend(fetched_data)



import os
from datetime import datetime, timedelta, timezone
from urllib.parse import urlparse

from biolib.api import client as api_client
from biolib.biolib_api_client.lfs_types import DataRecordVersion
from biolib.biolib_binary_format.utils import RemoteEndpoint
from biolib.biolib_logging import logger
from biolib.typing_utils import Optional


class DataRecordRemoteStorageEndpoint(RemoteEndpoint):
    def __init__(self, resource_version_uuid: str):
        self._resource_version_uuid: str = resource_version_uuid
        self._expires_at: Optional[datetime] = None
        self._presigned_url: Optional[str] = None

    def get_remote_url(self) -> str:
        if not self._presigned_url or not self._expires_at or datetime.now(timezone.utc) > self._expires_at:
            lfs_version: DataRecordVersion = api_client.get(
                path=f'/lfs/versions/{self._resource_version_uuid}/',
            ).json()

            app_caller_proxy_job_storage_base_url = os.getenv('BIOLIB_CLOUD_JOB_STORAGE_BASE_URL', '')
            if app_caller_proxy_job_storage_base_url:
                # Done to hit App Caller Proxy when downloading from inside an app
                parsed_url = urlparse(lfs_version['presigned_download_url'])
                self._presigned_url = f'{app_caller_proxy_job_storage_base_url}{parsed_url.path}?{parsed_url.query}'
            else:
                self._presigned_url = lfs_version['presigned_download_url']

            self._expires_at = datetime.now(timezone.utc) + timedelta(minutes=8)
            logger.debug(
                f'DataRecord "{self._resource_version_uuid}" fetched presigned URL '
                f'with expiry at {self._expires_at.isoformat()}'
            )

        return self._presigned_url



from .resource_types import ResourceDict
from .typing import TypedDict


class ResourcePermissionDict(TypedDict):
    uuid: str
    created_at: str
    action: str
    resource: ResourceDict
    target_resource: ResourceDict


class ResourcePermissionDetailedDict(ResourcePermissionDict):
    pass



# pylint: skip-file
"""
Code from:
https://github.com/gtsystem/python-remotezip
&
https://github.com/uktrade/stream-unzip
"""

import io
import zipfile

from functools import partial
from struct import Struct
import zlib

__all__ = ['RemoteIOError', 'RemoteZip']

from biolib._internal.http_client import HttpClient


class RemoteZipError(Exception):
    pass

class OutOfBound(RemoteZipError):
    pass


class RemoteIOError(RemoteZipError):
    pass


class RangeNotSupported(RemoteZipError):
    pass


class PartialBuffer:
    def __init__(self, buffer, offset, size, stream):
        self.buffer = buffer if stream else io.BytesIO(buffer.read())
        self.offset = offset
        self.size = size
        self.position = offset
        self.stream = stream

    def __repr__(self):
        return "<PartialBuffer off=%s size=%s stream=%s>" % (self.offset, self.size, self.stream)

    def read(self, size=0):
        if size == 0:
            size = self.offset + self.size - self.position

        content = self.buffer.read(size)
        self.position = self.offset + self.buffer.tell()
        return content

    def close(self):
        if not self.buffer.closed:
            self.buffer.close()
            if hasattr(self.buffer, 'release_conn'):
                self.buffer.release_conn()

    def seek(self, offset, whence):
        if whence == 2:
            self.position = self.size + self.offset + offset
        elif whence == 0:
            self.position = offset
        else:
            self.position += offset

        relative_position = self.position - self.offset

        if relative_position < 0 or relative_position >= self.size:
            raise OutOfBound("Position out of buffer bound")

        if self.stream:
            buff_pos = self.buffer.tell()
            if relative_position < buff_pos:
                raise OutOfBound("Negative seek not supported")

            skip_bytes = relative_position - buff_pos
            if skip_bytes == 0:
                return self.position
            self.buffer.read(skip_bytes)
        else:
            self.buffer.seek(relative_position)

        return self.position


class RemoteIO(io.IOBase):
    def __init__(self, fetch_fun, initial_buffer_size=64*1024):
        self.fetch_fun = fetch_fun
        self.initial_buffer_size = initial_buffer_size
        self.buffer = None
        self.file_size = None
        self.position = None
        self._seek_succeeded = False
        self.member_pos2size = None
        self._last_member_pos = None

    def set_pos2size(self, pos2size):
        self.member_pos2size = pos2size

    def read(self, size=0):
        if size == 0:
            size = self.file_size - self.buffer.position

        if not self._seek_succeeded:
            if self.member_pos2size is None:
                fetch_size = size
                stream = False
            else:
                try:
                    fetch_size = self.member_pos2size[self.buffer.position]
                    self._last_member_pos = self.buffer.position
                except KeyError:
                    if self._last_member_pos and self._last_member_pos < self.buffer.position:
                        fetch_size = self.member_pos2size[self._last_member_pos]
                        fetch_size -= (self.buffer.position - self._last_member_pos)
                    else:
                        raise OutOfBound("Attempt to seek outside boundary of current zip member")
                stream = True

            self._seek_succeeded = True
            self.buffer.close()
            self.buffer = self.fetch_fun((self.buffer.position, self.buffer.position + fetch_size -1), stream=stream)

        return self.buffer.read(size)

    def seekable(self):
        return True

    def seek(self, offset, whence=0):
        if whence == 2 and self.file_size is None:
            size = self.initial_buffer_size
            self.buffer = self.fetch_fun((-size, None), stream=False)
            self.file_size = self.buffer.size + self.buffer.position

        try:
            pos = self.buffer.seek(offset, whence)
            self._seek_succeeded = True
            return pos
        except OutOfBound:
            self._seek_succeeded = False
            return self.buffer.position   # we ignore the issue here, we will check if buffer is fine during read

    def tell(self):
        return self.buffer.position

    def close(self):
        if self.buffer:
            self.buffer.close()
            self.buffer = None


class RemoteZip(zipfile.ZipFile):
    def __init__(self, url, initial_buffer_size=64*1024):
        self.url = url

        rio = RemoteIO(self.fetch_fun, initial_buffer_size)
        super(RemoteZip, self).__init__(rio)
        rio.set_pos2size(self.get_position2size())

    def get_central_directory(self):
        return {
            file.filename: {
                attribute: getattr(file, attribute) for attribute in zipfile.ZipInfo.__slots__
            } for file in self.infolist()
        }

    def get_position2size(self):
        ilist = self.infolist()
        if len(ilist) == 0:
            return {}

        position2size = {ilist[-1].header_offset: self.start_dir - ilist[-1].header_offset}
        for i in range(len(ilist) - 1):
            m1, m2 = ilist[i: i+2]
            position2size[m1.header_offset] = m2.header_offset - m1.header_offset

        return position2size

    @staticmethod
    def make_buffer(io_buffer, content_range_header, stream):
        range_min, range_max = content_range_header.split("/")[0][6:].split("-")
        range_min, range_max = int(range_min), int(range_max)
        return PartialBuffer(io_buffer, range_min, range_max - range_min + 1, stream)

    @staticmethod
    def make_header(range_min, range_max):
        if range_max is None:
            return "bytes=%s%s" % (range_min, '' if range_min < 0 else '-')
        return "bytes=%s-%s" % (range_min, range_max)

    def fetch_fun(self, data_range, stream=False):
        range_header = self.make_header(*data_range)
        try:
            response = HttpClient.request(url=self.url, headers={'Range': range_header})
            if 'Content-Range' not in response.headers:
                raise RangeNotSupported("The server doesn't support range requests")

            return self.make_buffer(io.BytesIO(response.content), response.headers['Content-Range'], stream=False)
        except IOError as e:
            raise RemoteIOError(str(e))

    def stream_unzip(self, zipfile_chunks, password=None, chunk_size=65536):
        local_file_header_signature = b'\x50\x4b\x03\x04'
        local_file_header_struct = Struct('<H2sHHHIIIHH')
        zip64_version = 45
        zip64_compressed_size = 4294967295
        zip64_size_signature = b'\x01\x00'
        aes_extra_signature = b'\x01\x99'
        central_directory_signature = b'\x50\x4b\x01\x02'
        central_directory_info = self.get_central_directory()

        def next_or_truncated_error(it):
            try:
                return next(it)
            except StopIteration:
                raise TruncatedDataError from None

        def get_byte_readers(iterable):
            # Return functions to return/"replace" bytes from/to the iterable
            # - _yield_all: yields chunks as they come up (often for a "body")
            # - _get_num: returns a single `bytes` of a given length
            # - _return_unused: puts "unused" bytes "back", to be retrieved by a yield/get call

            chunk = b''
            offset = 0
            it = iter(iterable)

            def _yield_all():
                nonlocal chunk, offset

                while True:
                    if offset == len(chunk):
                        try:
                            chunk = next(it)
                        except StopIteration:
                            break
                        else:
                            offset = 0
                    to_yield = min(len(chunk) - offset, chunk_size)
                    offset = offset + to_yield
                    yield chunk[offset - to_yield:offset]

            def _yield_num(num):
                nonlocal chunk, offset

                while num:
                    if offset == len(chunk):
                        chunk = next_or_truncated_error(it)
                        offset = 0
                    to_yield = min(num, len(chunk) - offset, chunk_size)
                    offset = offset + to_yield
                    num -= to_yield
                    yield chunk[offset - to_yield:offset]

            def _get_num(num):
                return b''.join(_yield_num(num))

            def _return_unused(num_unused):
                nonlocal offset
                offset -= num_unused

            return _yield_all, _get_num, _return_unused

        def get_decompressor_none(num_bytes):
            num_decompressed = 0
            num_unused = 0

            def _decompress(compressed_chunk):
                nonlocal num_decompressed, num_unused
                to_yield = min(len(compressed_chunk), num_bytes - num_decompressed)
                num_decompressed += to_yield
                num_unused = len(compressed_chunk) - to_yield
                yield compressed_chunk[:to_yield]

            def _is_done():
                return num_decompressed == num_bytes

            def _num_unused():
                return num_unused

            return _decompress, _is_done, _num_unused

        def get_decompressor_deflate():
            dobj = zlib.decompressobj(wbits=-zlib.MAX_WBITS)

            def _decompress_single(compressed_chunk):
                try:
                    return dobj.decompress(compressed_chunk, chunk_size)
                except zlib.error as e:
                    raise DeflateError() from e

            def _decompress(compressed_chunk):
                uncompressed_chunk = _decompress_single(compressed_chunk)
                if uncompressed_chunk:
                    yield uncompressed_chunk

                while dobj.unconsumed_tail and not dobj.eof:
                    uncompressed_chunk = _decompress_single(dobj.unconsumed_tail)
                    if uncompressed_chunk:
                        yield uncompressed_chunk

            def _is_done():
                return dobj.eof

            def _num_unused():
                return len(dobj.unused_data)

            return _decompress, _is_done, _num_unused

        def get_decompressor_deflate64():
            uncompressed_chunks, is_done, num_bytes_unconsumed = stream_inflate64()

            def _decompress(compressed_chunk):
                yield from uncompressed_chunks((compressed_chunk,))

            return _decompress, is_done, num_bytes_unconsumed

        def yield_file(yield_all, get_num, return_unused):

            def get_flag_bits(flags):
                for b in flags:
                    for i in range(8):
                        yield (b >> i) & 1

            def parse_extra(extra):
                extra_offset = 0
                while extra_offset <= len(extra) - 4:
                    extra_signature = extra[extra_offset:extra_offset + 2]
                    extra_offset += 2
                    extra_data_size, = Struct('<H').unpack(extra[extra_offset:extra_offset + 2])
                    extra_offset += 2
                    extra_data = extra[extra_offset:extra_offset + extra_data_size]
                    extra_offset += extra_data_size
                    yield (extra_signature, extra_data)

            def get_extra_value(extra, if_true, signature, exception_if_missing, min_length, exception_if_too_short):
                if if_true:
                    try:
                        value = extra[signature]
                    except KeyError:
                        raise exception_if_missing()

                    if len(value) < min_length:
                        raise exception_if_too_short()
                else:
                    value = None

                return value

            def decrypt_weak_decompress(chunks, decompress, is_done, num_unused):
                key_0 = 305419896
                key_1 = 591751049
                key_2 = 878082192
                crc32 = zlib.crc32
                bytes_c = bytes

                def update_keys(byte):
                    nonlocal key_0, key_1, key_2
                    key_0 = ~crc32(bytes_c((byte,)), ~key_0) & 0xFFFFFFFF
                    key_1 = (key_1 + (key_0 & 0xFF)) & 0xFFFFFFFF
                    key_1 = ((key_1 * 134775813) + 1) & 0xFFFFFFFF
                    key_2 = ~crc32(bytes_c((key_1 >> 24,)), ~key_2) & 0xFFFFFFFF

                def decrypt(chunk):
                    chunk = bytearray(chunk)
                    for i, byte in enumerate(chunk):
                        temp = key_2 | 2
                        byte ^= ((temp * (temp ^ 1)) >> 8) & 0xFF
                        update_keys(byte)
                        chunk[i] = byte
                    return bytes(chunk)

                for byte in password:
                    update_keys(byte)

                if decrypt(get_num(12))[11] != mod_time >> 8:
                    raise IncorrectZipCryptoPasswordError()

                while not is_done():
                    yield from decompress(decrypt(next_or_truncated_error(chunks)))

                return_unused(num_unused())

            def decrypt_aes_decompress(chunks, decompress, is_done, num_unused, key_length_raw):
                try:
                    key_length, salt_length = {1: (16, 8), 2: (24, 12), 3: (32, 16)}[key_length_raw]
                except KeyError:
                    raise InvalidAESKeyLengthError(key_length_raw)

                salt = get_num(salt_length)
                password_verification_length = 2

                keys = PBKDF2(password, salt, 2 * key_length + password_verification_length, 1000)
                if keys[-password_verification_length:] != get_num(password_verification_length):
                    raise IncorrectAESPasswordError()

                decrypter = AES.new(
                    keys[:key_length], AES.MODE_CTR,
                    counter=Counter.new(nbits=128, little_endian=True)
                )
                hmac = HMAC.new(keys[key_length:key_length * 2], digestmod=SHA1)

                while not is_done():
                    chunk = next_or_truncated_error(chunks)
                    yield from decompress(decrypter.decrypt(chunk))
                    hmac.update(chunk[:len(chunk) - num_unused()])

                return_unused(num_unused())

                if get_num(10) != hmac.digest()[:10]:
                    raise HMACIntegrityError()

            def decrypt_none_decompress(chunks, decompress, is_done, num_unused):
                while not is_done():
                    yield from decompress(next_or_truncated_error(chunks))

                return_unused(num_unused())

            def get_crc_32_expected_from_data_descriptor(is_zip64, file_size_stored_as_long_long):
                dd_optional_signature = get_num(4)
                dd_so_far_num = \
                    0 if dd_optional_signature == b'PK\x07\x08' else \
                        4
                dd_so_far = dd_optional_signature[:dd_so_far_num]
                # Have to check both if zip64 and if we store as long long (8), since some zip64 store only as long (4)
                dd_remaining = \
                    (20 - dd_so_far_num) if is_zip64 and file_size_stored_as_long_long else \
                        (12 - dd_so_far_num)
                dd = dd_so_far + get_num(dd_remaining)
                crc_32_expected, = Struct('<I').unpack(dd[:4])
                return crc_32_expected

            def get_crc_32_expected_from_file_header():
                return crc_32_expected

            def read_data_and_crc_32_ignore(get_crc_32_expected, chunks):
                yield from chunks
                get_crc_32_expected()

            def read_data_and_crc_32_verify(get_crc_32_expected, chunks):
                crc_32_actual = zlib.crc32(b'')
                for chunk in chunks:
                    crc_32_actual = zlib.crc32(chunk, crc_32_actual)
                    yield chunk

                if crc_32_actual != get_crc_32_expected():
                    raise CRC32IntegrityError()

            version, flags, compression_raw, mod_time, mod_date, crc_32_expected, compressed_size_raw, uncompressed_size_raw, file_name_len, extra_field_len = \
                local_file_header_struct.unpack(get_num(local_file_header_struct.size))

            flag_bits = tuple(get_flag_bits(flags))
            if (
                    flag_bits[4]  # Enhanced deflating
                    or flag_bits[5]  # Compressed patched
                    or flag_bits[6]  # Strong encrypted
                    or flag_bits[13]  # Masked header values
            ):
                raise UnsupportedFlagsError(flag_bits)

            file_name = get_num(file_name_len)
            file_name_str = file_name.decode()

            # Get these attributes from central directory as they are incorrect in the File Header
            uncompressed_size_raw = central_directory_info[file_name_str]['file_size']
            extract_version = central_directory_info[file_name_str]['extract_version']
            central_directory_extra = central_directory_info[file_name_str]['extra']

            # Zip64 Extra field requires 20 bytes to store Header (2) + Field Length (2) + File size (8) + Compressed size (8)
            # The length of the filesize field determines if the length of the data descriptor is 12 or 20 bytes.
            if zip64_size_signature in central_directory_extra and len(central_directory_extra) >= 20:
                file_size_stored_as_long_long = True
            else:
                file_size_stored_as_long_long = False

            extra = dict(parse_extra(get_num(extra_field_len)))

            is_weak_encrypted = flag_bits[0] and compression_raw != 99
            is_aes_encrypted = flag_bits[0] and compression_raw == 99
            aes_extra = get_extra_value(extra, is_aes_encrypted, aes_extra_signature, MissingAESExtraError, 7,
                                        TruncatedAESExtraError)
            is_aes_2_encrypted = is_aes_encrypted and aes_extra[0:2] == b'\x02\x00'

            if is_weak_encrypted and password is None:
                raise MissingZipCryptoPasswordError()

            if is_aes_encrypted and password is None:
                raise MissingAESPasswordError()

            compression = \
                Struct('<H').unpack(aes_extra[5:7])[0] if is_aes_encrypted else \
                    compression_raw

            if compression not in (0, 8, 9):
                raise UnsupportedCompressionTypeError(compression)

            has_data_descriptor = flag_bits[3]
            is_zip64 = compressed_size_raw == zip64_compressed_size and uncompressed_size_raw == zip64_compressed_size \
                       or extract_version == zip64_version
            zip64_extra = get_extra_value(extra, not has_data_descriptor and is_zip64, zip64_size_signature,
                                          MissingZip64ExtraError, 16, TruncatedZip64ExtraError)

            # zip64_extra can be None in some cases where is_zip64 is True so it is necessary to check.
            uncompressed_size = \
                None if has_data_descriptor and compression in (8, 9) else \
                    Struct('<Q').unpack(zip64_extra[:8])[0] if is_zip64 and zip64_extra else \
                        uncompressed_size_raw

            decompressor = \
                get_decompressor_none(uncompressed_size) if compression == 0 else \
                    get_decompressor_deflate() if compression == 8 else \
                        get_decompressor_deflate64()

            decompressed_bytes = \
                decrypt_weak_decompress(yield_all(), *decompressor) if is_weak_encrypted else \
                    decrypt_aes_decompress(yield_all(), *decompressor,
                                           key_length_raw=aes_extra[4]) if is_aes_encrypted else \
                        decrypt_none_decompress(yield_all(), *decompressor)

            get_crc_32_expected = \
                partial(get_crc_32_expected_from_data_descriptor, is_zip64, file_size_stored_as_long_long) \
                    if has_data_descriptor else get_crc_32_expected_from_file_header

            crc_checked_bytes = \
                read_data_and_crc_32_ignore(get_crc_32_expected, decompressed_bytes) if is_aes_2_encrypted else \
                    read_data_and_crc_32_verify(get_crc_32_expected, decompressed_bytes)

            return file_name, uncompressed_size, crc_checked_bytes

        def all():
            yield_all, get_num, return_unused = get_byte_readers(zipfile_chunks)

            while True:
                signature = get_num(len(local_file_header_signature))
                if signature == local_file_header_signature:
                    yield yield_file(yield_all, get_num, return_unused)
                elif signature == central_directory_signature:
                    for _ in yield_all():
                        pass
                    break
                else:
                    raise UnexpectedSignatureError(signature)

        for file_name, file_size, unzipped_chunks in all():
            yield file_name, file_size, unzipped_chunks
            for _ in unzipped_chunks:
                raise UnfinishedIterationError()

class UnzipError(Exception):
    pass

class InvalidOperationError(UnzipError):
    pass

class UnfinishedIterationError(InvalidOperationError):
    pass

class UnzipValueError(UnzipError, ValueError):
    pass

class DataError(UnzipValueError):
    pass

class UncompressError(UnzipValueError):
    pass

class DeflateError(UncompressError):
    pass

class UnsupportedFeatureError(DataError):
    pass

class UnsupportedFlagsError(UnsupportedFeatureError):
    pass

class UnsupportedCompressionTypeError(UnsupportedFeatureError):
    pass

class TruncatedDataError(DataError):
    pass

class UnexpectedSignatureError(DataError):
    pass

class MissingExtraError(DataError):
    pass

class MissingZip64ExtraError(MissingExtraError):
    pass

class MissingAESExtraError(MissingExtraError):
    pass

class TruncatedExtraError(DataError):
    pass

class TruncatedZip64ExtraError(TruncatedExtraError):
    pass

class TruncatedAESExtraError(TruncatedExtraError):
    pass

class InvalidExtraError(TruncatedExtraError):
    pass

class InvalidAESKeyLengthError(TruncatedExtraError):
    pass

class IntegrityError(DataError):
    pass

class HMACIntegrityError(IntegrityError):
    pass

class CRC32IntegrityError(IntegrityError):
    pass

class PasswordError(UnzipValueError):
    pass

class MissingPasswordError(UnzipValueError):
    pass

class MissingZipCryptoPasswordError(MissingPasswordError):
    pass

class MissingAESPasswordError(MissingPasswordError):
    pass

class IncorrectPasswordError(PasswordError):
    pass

class IncorrectZipCryptoPasswordError(IncorrectPasswordError):
    pass

class IncorrectAESPasswordError(IncorrectPasswordError):
    pass



import re

from biolib._shared.types import Optional
from biolib._shared.types.resource import ResourceUriDict, SemanticVersionDict
from biolib.biolib_errors import BioLibError

URI_REGEX = re.compile(
    r'^(@(?P<resource_prefix>[\w._-]+)/)?'
    r'(?P<account_handle>[\w-]+)'
    r'(/(?P<resource_name>[\w-]+))?'
    r'(?::(?P<suffix>[^:]+))?$'
)
SEMVER_REGEX = re.compile(r'^(?P<major>0|[1-9]\d*)\.(?P<minor>0|[1-9]\d*)\.(?P<patch>0|[1-9]\d*)$')
TAG_REGEX = re.compile(r'^[a-z0-9-]{1,128}$')


def normalize_resource_name(string: str) -> str:
    return string.replace('-', '_').lower()


def parse_semantic_version(semantic_version: str) -> SemanticVersionDict:
    if match := SEMVER_REGEX.fullmatch(semantic_version):
        return SemanticVersionDict(
            major=int(match.group('major')),
            minor=int(match.group('minor')),
            patch=int(match.group('patch')),
        )

    raise ValueError('The version must be a valid semantic version in the format of major.minor.patch (1.2.3).')


def parse_resource_uri(uri: str, use_account_as_name_default: bool = True) -> ResourceUriDict:
    matches = URI_REGEX.match(uri)
    if matches is None:
        raise BioLibError(f"Could not parse resource uri '{uri}', uri did not match regex")

    version: Optional[SemanticVersionDict] = None
    tag: Optional[str] = None

    suffix = matches.group('suffix')
    if suffix and suffix != '*':
        try:
            version = parse_semantic_version(suffix)
        except ValueError:
            if TAG_REGEX.fullmatch(suffix):
                tag = suffix
            else:
                raise BioLibError(
                    f'Invalid version or tag "{suffix}". '
                    'Versions must be semantic versions like "1.2.3". '
                    'Tags must be lowercase alphanumeric or dashes and at most 128 characters.'
                ) from None

    resource_prefix_raw: Optional[str] = matches.group('resource_prefix')
    resource_prefix = resource_prefix_raw.lower() if resource_prefix_raw is not None else None
    account_handle: str = matches.group('account_handle')
    account_handle_normalized: str = normalize_resource_name(account_handle)
    resource_name: Optional[str] = matches.group('resource_name')

    if resource_name:
        resource_name_normalized = normalize_resource_name(resource_name)
    elif use_account_as_name_default:
        resource_name_normalized = account_handle_normalized
    else:
        resource_name_normalized = None

    return ResourceUriDict(
        resource_prefix=resource_prefix,
        account_handle=account_handle,
        account_handle_normalized=account_handle_normalized,
        resource_name_normalized=resource_name_normalized,
        resource_name=resource_name if resource_name is not None or not use_account_as_name_default else account_handle,
        version=version,
        tag=tag,
    )



from .typing import Optional, TypedDict


class SemanticVersionDict(TypedDict):
    major: int
    minor: int
    patch: int


class ResourceUriDict(TypedDict):
    account_handle_normalized: str
    account_handle: str
    resource_name_normalized: Optional[str]
    resource_name: Optional[str]
    resource_prefix: Optional[str]
    version: Optional[SemanticVersionDict]
    tag: Optional[str]



import enum

from biolib.typing_utils import Any, TypedDict


class Proxy(TypedDict):
    container: Any
    hostname: str
    ip: str


class DockerDiffKind(enum.Enum):
    CHANGED = 0
    ADDED = 1
    DELETED = 2



from biolib.typing_utils import TypedDict


class DataRecordVersion(TypedDict):
    presigned_download_url: str
    size_bytes: int
    uri: str
    uuid: str


class DataRecordInfo(TypedDict):
    uri: str
    uuid: str


class DataRecordVersionInfo(TypedDict):
    resource_uri: str
    resource_uuid: str
    resource_version_uuid: str



import base64
import io
import ipaddress
import tarfile
import time
from urllib.parse import urlparse

from docker.models.containers import Container
from docker.models.networks import Network
from docker.types import EndpointConfig

from biolib import utils
from biolib.biolib_api_client import BiolibApiClient, RemoteHost
from biolib.biolib_docker_client import BiolibDockerClient
from biolib.biolib_errors import BioLibError
from biolib.biolib_logging import logger_no_user_data
from biolib.compute_node.cloud_utils import CloudUtils
from biolib.compute_node.utils import BIOLIB_PROXY_NETWORK_NAME
from biolib.compute_node.webserver.proxy_utils import get_biolib_nginx_proxy_image
from biolib.typing_utils import Dict, List, Optional


def get_static_ip_from_network(network: Network, offset: int = 2) -> str:
    ipam_config = network.attrs['IPAM']['Config']
    if not ipam_config:
        raise BioLibError(f'Network {network.name} has no IPAM configuration')

    subnet_str = ipam_config[0]['Subnet']
    subnet = ipaddress.ip_network(subnet_str, strict=False)

    static_ip = str(subnet.network_address + offset)

    return static_ip


# Prepare for remote hosts with specified port
class RemoteHostExtended(RemoteHost):
    ports: List[int]


class RemoteHostMapping:
    def __init__(self, hostname: str, ports: List[int], network: Network, static_ip: str):
        self.hostname = hostname
        self.ports = ports
        self.network = network
        self.static_ip = static_ip


class RemoteHostProxy:
    def __init__(
        self,
        remote_host_mappings: List[RemoteHostMapping],
        job_id: str,
        app_caller_network: Optional[Network] = None,
    ):
        self._remote_host_mappings = remote_host_mappings
        self._app_caller_network = app_caller_network
        self.is_app_caller_proxy = app_caller_network is not None

        if not job_id:
            raise Exception('RemoteHostProxy missing argument "job_id"')

        suffix = '-AppCallerProxy' if app_caller_network else ''
        self._name = f'biolib-remote-host-proxy-{job_id}{suffix}'
        self._job_uuid = job_id
        self._container: Optional[Container] = None
        self._docker = BiolibDockerClient().get_docker_client()

    def get_hostname_to_ip_mapping(self) -> Dict[str, str]:
        return {mapping.hostname: mapping.static_ip for mapping in self._remote_host_mappings}

    def get_remote_host_networks(self) -> List[Network]:
        networks = [mapping.network for mapping in self._remote_host_mappings]
        return networks

    def get_ip_address_on_network(self, network: Network) -> str:
        if not self._container:
            raise BioLibError('RemoteHostProxy not yet started')

        container_networks = self._container.attrs['NetworkSettings']['Networks']
        if network.name in container_networks:
            ip_address: str = container_networks[network.name]['IPAddress']
            if not ip_address:
                raise BioLibError(f'No IP address found for network {network.name}')
            return ip_address

        raise BioLibError(f'RemoteHostProxy not connected to network {network.name}')

    def start(self) -> None:
        docker = BiolibDockerClient.get_docker_client()

        networking_config: Optional[Dict[str, EndpointConfig]] = (
            None
            if not self.is_app_caller_proxy
            else {
                BIOLIB_PROXY_NETWORK_NAME: docker.api.create_endpoint_config(
                    aliases=[f'biolib-app-caller-proxy-{self._job_uuid}']
                )
            }
        )

        for index in range(3):
            logger_no_user_data.debug(f'Attempt {index} at creating RemoteHostProxy container "{self._name}"...')
            try:
                self._container = docker.containers.create(
                    detach=True,
                    image=get_biolib_nginx_proxy_image(),
                    name=self._name,
                    network=BIOLIB_PROXY_NETWORK_NAME,
                    networking_config=networking_config,
                )
                break
            except Exception as error:
                logger_no_user_data.exception(f'Failed to create container "{self._name}" hit error: {error}')

            logger_no_user_data.debug('Sleeping before re-trying container creation...')
            time.sleep(3)

        if not self._container or not self._container.id:
            raise BioLibError(f'Exceeded re-try limit for creating container {self._name}')

        for mapping in self._remote_host_mappings:
            mapping.network.connect(self._container.id, ipv4_address=mapping.static_ip)
            logger_no_user_data.debug(
                f'Connected proxy to network {mapping.network.name} with static IP {mapping.static_ip}'
            )

        if self._app_caller_network:
            self._app_caller_network.connect(self._container.id)
            logger_no_user_data.debug(f'Connected app caller proxy to network {self._app_caller_network.name}')

        self._write_nginx_config_to_container()

        self._container.start()

        proxy_is_ready = False
        for retry_count in range(1, 5):
            time.sleep(0.5 * retry_count)
            # Use the container logs as a health check.
            # By using logs instead of a http endpoint on the NGINX we avoid publishing a port of container to the host
            container_logs = self._container.logs()
            if b'ready for start up\n' in container_logs or b'start worker process ' in container_logs:
                proxy_is_ready = True
                break

        if not proxy_is_ready:
            self.terminate()
            raise Exception('RemoteHostProxy did not start properly')

        self._container.reload()

    def terminate(self):
        # TODO: Implement nice error handling in this method

        if self._container:
            self._container.remove(force=True)

    def _write_nginx_config_to_container(self) -> None:
        if not self._container:
            raise Exception('RemoteHostProxy container not defined when attempting to write NGINX config')

        docker = BiolibDockerClient.get_docker_client()
        upstream_hostname = urlparse(BiolibApiClient.get().base_url).hostname
        if self.is_app_caller_proxy:
            if not utils.IS_RUNNING_IN_CLOUD:
                raise BioLibError('Calling apps inside apps is not supported in local compute environment')

            logger_no_user_data.debug(f'Job "{self._job_uuid}" writing config for and starting App Caller Proxy...')
            config = CloudUtils.get_webserver_config()
            compute_node_uuid = config['compute_node_info']['public_id']
            compute_node_auth_token = config['compute_node_info']['auth_token']

            # TODO: Get access_token from new API class instead
            access_token = BiolibApiClient.get().access_token
            bearer_token = f'Bearer {access_token}' if access_token else ''

            biolib_index_basic_auth = f'compute_node|admin:{compute_node_auth_token},{self._job_uuid}'
            biolib_index_basic_auth_base64 = base64.b64encode(biolib_index_basic_auth.encode('utf-8')).decode('utf-8')

            nginx_config = f"""
events {{
  worker_connections  1024;
}}

http {{
    client_max_body_size 0;
    map $request_method $bearer_token_on_post {{
        POST       "{bearer_token}";
        default    "";
    }}

    map $request_method $bearer_token_on_get {{
        GET        "{bearer_token}";
        default    "";
    }}

    map $request_method $bearer_token_on_patch_and_get {{
        PATCH      "{bearer_token}";
        GET        "{bearer_token}";
        default    "";
    }}

    server {{
        listen       80;
        resolver 127.0.0.11 ipv6=off valid=30s;
        set $upstream_hostname {upstream_hostname};

        location ~* "^/api/jobs/cloud/(?<job_id>[a-z0-9-]{{36}})/status/$" {{
            proxy_pass               https://$upstream_hostname/api/jobs/cloud/$job_id/status/;
            proxy_set_header         authorization $bearer_token_on_get;
            proxy_set_header         cookie "";
            proxy_ssl_server_name    on;
        }}

        location ~* "^/api/jobs/cloud/$" {{
            # Note: Using $1 here as URI part from regex must be used for proxy_pass
            proxy_pass               https://$upstream_hostname/api/jobs/cloud/$1;
            proxy_set_header         authorization $bearer_token_on_post;
            proxy_set_header         cookie "";
            proxy_ssl_server_name    on;
        }}

        location ~* "^/api/jobs/(?<job_id>[a-z0-9-]{{36}})/storage/input/start_upload/$" {{
            proxy_pass               https://$upstream_hostname/api/jobs/$job_id/storage/input/start_upload/;
            proxy_set_header         authorization "";
            proxy_set_header         cookie "";
            proxy_ssl_server_name    on;
        }}

        location ~* "^/api/jobs/(?<job_id>[a-z0-9-]{{36}})/storage/input/presigned_upload_url/$" {{
            proxy_pass               https://$upstream_hostname/api/jobs/$job_id/storage/input/presigned_upload_url/$is_args$args;
            proxy_set_header         authorization "";
            proxy_set_header         cookie "";
            proxy_ssl_server_name    on;
        }}

        location ~* "^/api/jobs/(?<job_id>[a-z0-9-]{{36}})/storage/input/complete_upload/$" {{
            proxy_pass               https://$upstream_hostname/api/jobs/$job_id/storage/input/complete_upload/;
            proxy_set_header         authorization "";
            proxy_set_header         cookie "";
            proxy_ssl_server_name    on;
        }}

        location ~* "^/api/jobs/(?<job_id>[a-z0-9-]{{36}})/main_result/$" {{
            proxy_pass                  https://$upstream_hostname/api/jobs/$job_id/main_result/;
            proxy_set_header            authorization "";
            proxy_set_header            cookie "";
            proxy_pass_request_headers  on;
            proxy_ssl_server_name       on;
        }}

        location ~* "^/api/jobs/(?<job_id>[a-z0-9-]{{36}})/$" {{
            proxy_pass               https://$upstream_hostname/api/jobs/$job_id/;
            proxy_set_header         authorization $bearer_token_on_patch_and_get;
            proxy_set_header         caller-job-uuid "{self._job_uuid}";
            proxy_set_header         cookie "";
            proxy_ssl_server_name    on;
        }}

        location ~* "^/api/jobs/create_job_with_data/$" {{
            # Note: Using $1 here as URI part from regex must be used for proxy_pass
            proxy_pass               https://$upstream_hostname/api/jobs/create_job_with_data/$1;
            proxy_set_header         authorization $bearer_token_on_post;
            proxy_set_header         caller-job-uuid "{self._job_uuid}";
            proxy_set_header         cookie "";
            proxy_ssl_server_name    on;
        }}

        location ~* "^/api/jobs/$" {{
            # Note: Using $1 here as URI part from regex must be used for proxy_pass
            proxy_pass               https://$upstream_hostname/api/jobs/$1;
            proxy_set_header         authorization $bearer_token_on_post;
            proxy_set_header         caller-job-uuid "{self._job_uuid}";
            proxy_set_header         cookie "";
            proxy_ssl_server_name    on;
        }}

        location ~ "^/api/jobs/{self._job_uuid}/notes/$" {{
            # Note: Using $1 here as URI part from regex must be used for proxy_pass
            proxy_pass               https://$upstream_hostname/api/jobs/{self._job_uuid}/notes/$1;
            proxy_set_header         authorization "";
            proxy_set_header         job-auth-token "";
            proxy_set_header         compute-node-auth-token "{compute_node_auth_token}";
            proxy_set_header         compute-node-uuid "{compute_node_uuid}";
            proxy_set_header         cookie "";
            proxy_ssl_server_name    on;
        }}

        location ~ "^/api/auth/oauth-token-exchange/$" {{
            # Note: Using $1 here as URI part from regex must be used for proxy_pass
            proxy_pass               https://$upstream_hostname/api/auth/oauth-token-exchange/$1;
            proxy_set_header         authorization "";
            proxy_set_header         compute-node-auth-token "{compute_node_auth_token}";
            proxy_set_header         job-uuid "{self._job_uuid}";
            proxy_set_header         cookie "";
            proxy_ssl_server_name    on;
        }}

        location /api/lfs/ {{
            proxy_pass               https://$upstream_hostname$request_uri;
            proxy_set_header         authorization "";
            proxy_set_header         compute-node-auth-token "{compute_node_auth_token}";
            proxy_set_header         job-uuid "{self._job_uuid}";
            proxy_set_header         cookie "";
            proxy_ssl_server_name    on;
        }}

        location /api/app/ {{
            proxy_pass               https://$upstream_hostname$request_uri;
            proxy_set_header         authorization "";
            proxy_set_header         compute-node-auth-token "{compute_node_auth_token}";
            proxy_set_header         job-uuid "{self._job_uuid}";
            proxy_set_header         cookie "";
            proxy_ssl_server_name    on;
        }}

        location /api/resources/data-records/ {{
            proxy_pass               https://$upstream_hostname$request_uri;
            proxy_set_header         authorization "";
            proxy_set_header         compute-node-auth-token "{compute_node_auth_token}";
            proxy_set_header         job-uuid "{self._job_uuid}";
            proxy_set_header         cookie "";
            proxy_ssl_server_name    on;
        }}

        location /api/resources/versions/ {{
            proxy_pass               https://$upstream_hostname$request_uri;
            proxy_set_header         authorization "";
            proxy_set_header         compute-node-auth-token "{compute_node_auth_token}";
            proxy_set_header         job-uuid "{self._job_uuid}";
            proxy_set_header         cookie "";
            proxy_ssl_server_name    on;
        }}

        location /api/proxy/files/data-record-versions/ {{
            proxy_pass               https://$upstream_hostname$request_uri;
            proxy_set_header         authorization "";
            proxy_set_header         compute-node-auth-token "{compute_node_auth_token}";
            proxy_set_header         job-uuid "{self._job_uuid}";
            proxy_set_header         cookie "";
            proxy_ssl_server_name    on;
        }}

        location /api/proxy/index/ {{
            proxy_pass               https://$upstream_hostname$request_uri;
            proxy_set_header         authorization "Basic {biolib_index_basic_auth_base64}";
            proxy_set_header         cookie "";
            proxy_ssl_server_name    on;
        }}

        location ~* "^/api/accounts/(?<account_id>[a-z0-9-]{{36}})/metrics/jobs/$" {{
            proxy_pass               https://$upstream_hostname/api/accounts/$account_id/metrics/jobs/$is_args$args;
            proxy_set_header         authorization "";
            proxy_set_header         compute-node-auth-token "{compute_node_auth_token}";
            proxy_set_header         job-uuid "{self._job_uuid}";
            proxy_set_header         cookie "";
            proxy_ssl_server_name    on;
        }}

        location /api/ {{
            proxy_pass               https://$upstream_hostname$request_uri;
            proxy_set_header         authorization "";
            proxy_set_header         cookie "";
            proxy_ssl_server_name    on;
        }}

        location /proxy/storage/job-storage/ {{
            proxy_pass               https://$upstream_hostname$request_uri;
            proxy_set_header         authorization "";
            proxy_set_header         cookie "";
            proxy_ssl_server_name    on;
        }}

        location /proxy/storage/lfs/versions/ {{
            proxy_pass               https://$upstream_hostname$request_uri;
            proxy_set_header         authorization "";
            proxy_set_header         cookie "";
            proxy_ssl_server_name    on;
        }}

        location /proxy/cloud/ {{
            proxy_pass               https://$upstream_hostname$request_uri;
            proxy_set_header         authorization "";
            proxy_set_header         cookie "";
            proxy_ssl_server_name    on;
        }}

        location /proxy/index/ {{
            proxy_pass               https://$upstream_hostname$request_uri;
            proxy_set_header         authorization "Basic {biolib_index_basic_auth_base64}";
            proxy_set_header         cookie "";
            proxy_ssl_server_name    on;
        }}

        location / {{
            return 404 "Not found";
        }}
    }}

    server {{
        listen       1080;
        resolver 127.0.0.11 ipv6=off valid=30s;

        if ($http_biolib_result_uuid != "{self._job_uuid}") {{
            return 403 "Invalid or missing biolib-result-uuid header";
        }}

        if ($http_biolib_result_port = "") {{
            return 400 "Missing biolib-result-port header";
        }}

        location / {{
            proxy_pass               http://main:$http_biolib_result_port$request_uri;
            proxy_set_header         biolib-result-uuid "";
            proxy_set_header         biolib-result-port "";
            proxy_pass_request_headers on;
        }}
    }}
}}
"""
        else:
            port_to_mappings: Dict[int, List[RemoteHostMapping]] = {}
            for mapping in self._remote_host_mappings:
                for port in mapping.ports:
                    if port not in port_to_mappings:
                        port_to_mappings[port] = []
                    port_to_mappings[port].append(mapping)

            nginx_config = """
events {}
error_log /dev/stdout info;
stream {
    resolver 127.0.0.11 valid=30s;"""

            for port, mappings in port_to_mappings.items():
                nginx_config += f"""
    map $server_addr $backend_{port} {{"""
                for mapping in mappings:
                    nginx_config += f'\n        {mapping.static_ip} {mapping.hostname}:{port};'

                nginx_config += f"""
    }}
    server {{
        listen 0.0.0.0:{port};
        proxy_pass $backend_{port};
    }}
    server {{
        listen 0.0.0.0:{port} udp;
        proxy_pass $backend_{port};
    }}"""

            nginx_config += """
}
"""

        nginx_config_bytes = nginx_config.encode()
        tarfile_in_memory = io.BytesIO()
        with tarfile.open(fileobj=tarfile_in_memory, mode='w:gz') as tar:
            info = tarfile.TarInfo('/nginx.conf')
            info.size = len(nginx_config_bytes)
            tar.addfile(info, io.BytesIO(nginx_config_bytes))

        tarfile_bytes = tarfile_in_memory.getvalue()
        tarfile_in_memory.close()
        docker.api.put_archive(self._container.id, '/etc/nginx', tarfile_bytes)



import json
from typing import Any, Dict

from biolib import api
from biolib._index.types import IndexInfo
from biolib._internal.index import get_index_from_uri
from biolib.biolib_api_client import BiolibApiClient
from biolib.biolib_logging import logger


class Index:
    def __init__(self, _internal_state: IndexInfo):
        self._state = _internal_state

    def __repr__(self) -> str:
        return f'Index: {self._state["resource_uri"]}'

    @property
    def uri(self) -> str:
        return self._state['resource_uri']

    @property
    def id(self) -> str:
        return f"{self._state['group_uuid']}.{self._state['resource_uuid']}".replace("-", "_")

    @staticmethod
    def get_by_uri(uri: str) -> 'Index':
        return Index(_internal_state=get_index_from_uri(uri))

    @staticmethod
    def create(uri: str, config: Dict[str, Any]) -> str:
        BiolibApiClient.assert_is_signed_in(authenticated_action_description='create an Index')

        response = api.client.post(
            path='/resources/indexes/',
            data={
                'uri': uri,
                'index_config': config,
            },
        )
        result = response.json()
        created_uri: str = result['uri']
        logger.info(f"Successfully created Index '{created_uri}'")
        return created_uri

    @staticmethod
    def create_from_config_file(uri: str, config_path: str) -> str:
        with open(config_path) as config_file:
            index_config = json.load(config_file)

        return Index.create(uri=uri, config=index_config)



from .typing import TypedDict


class FileZipMetadataDict(TypedDict):
    header_start: int
    size_on_disk: int


class FileNodeDict(TypedDict):
    dir_path: str
    is_dir: bool
    name: str
    size: int


class ZipFileNodeDict(FileNodeDict):
    zip_meta: FileZipMetadataDict
